@@ -2,10 +2,13 @@ from .base_parquet_artifact import BaseParquetArtifact
 from .base_data_cube import BaseDataCube
 from .base_attacher import make_attacher
 from .base_parquet_reader import BaseParquetReader
+from .hybrid_data_loader import HybridDataLoader
+
 __all__ = [
     "BaseDataCube",
     "BaseParquetArtifact",
     "make_attacher",
-    "BaseParquetReader"
+    "BaseParquetReader",
+    "HybridDataLoader",
 ]
 
